from meterviewer.datasets import imgv
from meterviewer.views import np_dataset
from tests.utils import show_img


def test_view_merge_np(root_path):
  """test view_merge_np"""
  d = root_path / "generated"
  np_dataset.view_merge_np(d)
  assert True


def test_view_imgv(root_path):
  """test view_imgv"""
  img_path = root_path / "lens_6/XL/XL/M1L3XL/2018-11-23-12-16-01.jpg"
  # assert False, img_path
  im, v, rect = imgv.view_one_img_v(img_path)
  show_img(im)
  assert v == "000994"



from meterviewer.validator import check_if_meterdataset

dataset_path = "/data/xiu-hao/work/Dataset/MeterData/lens_6/XL/XL/M10L1XL"


def test_check_if_meterdataset():
  check_if_meterdataset(dataset_path)



import pathlib

import pytest

from meterviewer.generator.jsondb import set_local_config


@pytest.fixture
def set_config():
  config_path = pathlib.Path(__file__).parent / "config.toml"
  set_local_config(config_path)
  return config_path



import cv2
import numpy as np


def four_point_transform(image, pts):
  # 获取输入坐标点
  rect = np.float32(pts)

  # 计算输入的w和h值
  widthA = np.sqrt(((rect[2][0] - rect[3][0]) ** 2) + ((rect[2][1] - rect[3][1]) ** 2))
  widthB = np.sqrt(((rect[1][0] - rect[0][0]) ** 2) + ((rect[1][1] - rect[0][1]) ** 2))
  maxWidth = max(int(widthA), int(widthB))

  heightA = np.sqrt(((rect[1][0] - rect[2][0]) ** 2) + ((rect[1][1] - rect[2][1]) ** 2))
  heightB = np.sqrt(((rect[0][0] - rect[3][0]) ** 2) + ((rect[0][1] - rect[3][1]) ** 2))
  maxHeight = max(int(heightA), int(heightB))

  # 变换后的坐标
  dst = np.array(
    [[0, 0], [maxWidth - 1, 0], [maxWidth - 1, maxHeight - 1], [0, maxHeight - 1]],
    dtype="float32",
  )

  # 计算变换矩阵
  M = cv2.getPerspectiveTransform(rect, dst)

  # 执行变换
  warped = cv2.warpPerspective(image, M, (maxWidth, maxHeight))

  return warped


def draw_points_and_lines(image, points):
  # 创建图片副本，避免修改原图
  img_copy = image.copy()

  # 绘制点
  for i, point in enumerate(points):
    x, y = map(int, point)
    # 画点（圆形），参数：图片，中心点，半径，颜色，粗细（-1表示填充）
    cv2.circle(img_copy, (x, y), 5, (0, 0, 255), -1)  # 红色点
    # 在点旁边添加序号
    cv2.putText(
      img_copy,
      str(i + 1),
      (x + 10, y + 10),
      cv2.FONT_HERSHEY_SIMPLEX,
      0.8,
      (0, 255, 0),
      2,
    )

  # 连接点（可选）
  # 连接成四边形
  for i in range(len(points)):
    pt1 = tuple(map(int, points[i]))
    pt2 = tuple(map(int, points[(i + 1) % len(points)]))
    cv2.line(img_copy, pt1, pt2, (255, 0, 0), 2)  # 蓝色线

  return img_copy



# compare two image
import hashlib

from meterviewer import types as T


def comp_ims(im1: T.NpImage, im2: T.NpImage) -> bool:
  """对比图片内容看是否完全一致"""
  return get_hash(im1) == get_hash(im2)


def get_hash(img: T.NpImage) -> bytes:
  return hashlib.md5(img.tobytes()).digest()



import pytest
import pathlib
import typing as t
from pathlib import Path as P
from meterviewer.config import get_root_path


@pytest.fixture
def alldata() -> P:
  return P("alldata.db")


@pytest.fixture
def remove_files() -> t.Generator[t.Callable[[P], None], None, None]:
  filepath_list = []

  def func(filep: P):
    filepath_list.append(filep)

  yield func
  for filepath in filepath_list:
    filepath.unlink()


@pytest.fixture
def root_path() -> pathlib.Path:
  return get_root_path()



from __future__ import annotations
from meterviewer.models.littledb import get_session, Item
from pathlib import Path as P
from sqlalchemy import ScalarResult, select, text, update


def get_first_item(dbpath: P) -> Item:
  session = get_session(str(dbpath))
  stmt = select(Item).where(Item.id == 1)
  result = session.execute(stmt)
  return result.scalar_one()


def get_carry_items(dbpath: P | str) -> ScalarResult[Item]:
  # get items with is_carry == 1
  session = get_session(dbpath)
  stmt = select(Item).where(Item.is_carry == 1)
  result = session.execute(stmt)
  return result.scalars()


def update_schema(dbpath: P):
  # add column carry_num to Item
  session = get_session(str(dbpath))
  session.execute(text("ALTER TABLE Item ADD COLUMN carry_num INTEGER"))
  session.commit()


def update_item(dbpath: P, id: int, carry_num: int):
  session = get_session(dbpath)
  if carry_num != 0:
    session.execute(update(Item).where(Item.id == id).values(carry_num=carry_num))
  else:
    session.execute(
      update(Item).where(Item.id == id).values(carry_num=carry_num, is_carry=0)
    )
  session.commit()



import glob
import pathlib

import cv2
import matplotlib.pyplot as plt

from meterviewer.datasets.read.config import get_xml_config
from meterviewer.datasets.read.detection import read_area_pos
from meterviewer.img.draw import draw_rectangle


class MeterSet(object):
  def __init__(self, root_path: pathlib.Path, name: str):
    self.name = name
    self.root_path = root_path
    self.image_list: list[str] = []
    self.load_list()

  def images(self, i: int):
    if i > len(self.image_list):
      raise ValueError(f"index {i} out of range")
    return cv2.imread(self.image_list[i])

  def print_img(self, i: int, with_area: bool = False):
    img = self.images(i)
    if with_area:
      rect = self.pos(i)
      img = draw_rectangle(img, rect)
    plt.imshow(img)
    plt.show()

  def __len__(self):
    return len(self.image_list)

  def values(self, i: int):
    if i > len(self.image_list):
      raise ValueError(f"index {i} out of range")
    v, _ = get_xml_config(pathlib.Path(self.image_list[i]))
    return v

  def pos(self, i: int):
    if i > len(self.image_list):
      raise ValueError(f"index {i} out of range")
    filepath = self.image_list[i]
    rect = read_area_pos(pathlib.Path(filepath))
    return rect

  def load_list(self):
    self.image_list = glob.glob(str(self.root_path / self.name / "*.jpg"))



import cv2

from meterviewer import types as T


# draw rectangle on image
def draw_rectangle(im: T.NpImage, rect: T.Rect) -> T.NpImage:
  """draw rectangle on image"""
  x0, y0, x1, y1 = rect.xmin, rect.ymin, rect.xmax, rect.ymax
  writable_im = im.copy()
  cv2.rectangle(writable_im, (x0, y0), (x1, y1), (0, 255, 0), 2)
  return writable_im


# write text on image
def draw_text(
  im: T.NpImage,
  text: str,
) -> T.NpImage:
  """write text on image"""
  x, y = 0, 30
  writable_im = im.copy()
  cv2.putText(writable_im, text, (x, y), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0), 2)
  return writable_im



"""
Using MeterSet to construct torch dataset

Args:
  root_dir: root directory of the dataset
  stage: train or test
  train_folder: train folder name
  train_list: train list
"""

import pathlib
import warnings

try:
  from torch.utils.data import Dataset
except ImportError:
  warnings.warn(
    "torch is not installed. Please install it with `pip install torch`.", stacklevel=2
  )

  class Dataset:
    pass


from meterviewer.meterset import MeterSet


class MyMeterSet(MeterSet):
  def __str__(self):
    return f"MeterSet(name={self.name}, root_path={self.root_path})"

  def __len__(self):
    return len(self.image_list)


class MeterDataset(Dataset):
  def __init__(
    self,
    root_dir,
    stage: str,
    transform=None,
  ):
    self.root_dir = root_dir
    self.stage = stage
    self.transform = transform
    self.load_metersets()
    self.length = sum([self.get_len(meterset) for meterset in self.metersets])
    self.start_end_list = self.scan()

  def load_metersets(self):
    raise NotImplementedError("Please implement this method")

  def build_metersets(
    self,
    train_list: list[str],
    train_folder: str,
    test_list: list[str],
    test_folder: str,
  ):
    if self.stage == "train":
      self.metersets = [
        MyMeterSet(name=name, root_path=pathlib.Path(self.root_dir) / train_folder)
        for name in train_list
      ]
    elif self.stage == "test":
      self.metersets = [
        MyMeterSet(name=name, root_path=pathlib.Path(self.root_dir) / test_folder)
        for name in test_list
      ]

  def get_len(self, meterset: MeterSet):
    return len(meterset)

  def __len__(self):
    return self.length

  def check(self):
    for meterset in self.metersets:
      assert len(meterset.image_list) > 0, meterset

  def get_item(self, meterset, index):
    """get item from one metersets"""
    sample = {
      "image": meterset.images(index),
      "values": meterset.values(index),
      "pos": meterset.pos(index),
    }
    if self.transform is not None:
      sample = self.transform(sample)
    return sample

  def scan(self):
    """扫描 metersets，返回一个开始结束索引的列表，__getitem__ 可以借助这个列表来快速定位"""
    start_end_list = []
    for i, meterset in enumerate(self.metersets):
      start_end_list.append(
        (
          sum(self.get_len(meterset) for meterset in self.metersets[:i]),
          sum(self.get_len(meterset) for meterset in self.metersets[: i + 1]),
        )
      )
    return start_end_list

  def __getitem__(self, index):
    """get item from all over the metersets"""
    for i, (start, end) in enumerate(self.start_end_list):
      if start <= index < end:
        return self.get_item(self.metersets[i], index - start)
    raise IndexError(f"index {index} out of range")



"""根据数值判断进位情况；与数值相关的函数"""

import typing as t
from . import types as T


def is_carry(val: str) -> bool:
  return val[-1] in ("0", "9")


def is_carry_over(digit: T.DigitInt) -> bool:
  if digit[-1] == 0 or digit[-1] == 9:
    return True
  return False


def make_full(arr: t.List, length: int):
  new_arr = arr.copy()
  if len(arr) <= length:
    new_arr = [False] * (length - len(arr)) + new_arr
  else:
    raise ValueError(f"length is too short, len(arr) == {len(arr)}")
  return new_arr


def get_carry_array(digit: T.DigitStr) -> t.List[bool]:
  """get digit array from given digit string"""
  assert len(digit) >= 5
  res = []
  # 不是进位
  if not is_carry("".join(digit)):
    return make_full(res, len(digit))
  else:
    res.append(True)  # last digit is 0 or 9
    for c in reversed(digit[:-1]):
      res.append(True)
      if not c == "9":  # not continue when not 9
        break

  res = make_full(res, len(digit))
  return res


# def label_check(label: T.Label):
#     pass



from __future__ import annotations
import typing as t

import numpy as np
from PIL import Image

from meterviewer import types as T


def resize_img(img: T.NpImage, size: T.ImgSize) -> T.NpImage:
  # resize will reverse the size. height -> weight, weight -> height
  im = np.asarray(Image.fromarray(img).resize(list(reversed(size))), dtype=np.uint8)
  assert list(im.shape[:2]) == list(size), (im.shape, size)
  return im


def resize_imglist(
  imglist: T.ImgList,
  size: t.Optional[T.ImgSize] = None,
) -> T.ImgList:
  if not size:
    size = list(imglist[0].shape[:2])
  return [resize_img(img, size) for img in imglist]


def check_img_size(
  img: T.NpImage, size: t.Tuple[int, int], then: t.Callable[[T.NpImage], t.Any]
) -> bool | t.Any:
  if img.shape == size:
    return then(img)
  return False


def size_check(img_list: t.List[T.NpImage], size: t.Optional[t.List[int]] = None):
  assert img_list != [], "img_list should not be empty"
  if not size:
    size = list(img_list[0].shape)

  for i, img in enumerate(img_list):
    if list(img.shape) != size:
      raise ValueError(f"image: {i} size: {img.shape}, not match size: {size}")



from meterviewer.img import process

# import functools
# from matplotlib import pyplot as plt
from tests.utils import show_img, gen_img


def test_resize_imglist():
  imglist = [gen_img(size=(35, 25, 3)), gen_img(size=(34, 25, 3))]
  process.resize_imglist(imglist, size=[35, 25])

  imglist = [gen_img(size=(31, 170, 3)), gen_img(size=(31, 175, 3))]
  process.resize_imglist(imglist)


def test_resize_img():
  im = gen_img(size=(35, 25, 3))
  process.resize_img(im, size=[35, 25])


def test_number_to_string():
  assert process.number_to_string(1, 5) == list("00001")
  assert process.number_to_string(10, 5) == ["0", "0", "0", "1", "0"]
  assert process.number_to_string(100, 5) == list("00100")
  assert process.number_to_string(1000, 5) == list("01000")


def test_join():
  imglist = [
    process.get_random_img(1, process.img_from),
    process.get_random_img(2, process.img_from),
    process.get_random_img(3, process.img_from),
  ]
  res = process.join_img(imglist, process.empty_check)
  assert res.shape[1] > res.shape[0]
  # seems correct
  show_img(res)



import typing as t

global_func: t.Dict[str, t.Any] = {}


def use_func(name: str):
  f = global_func.get(name, None)
  if not f:
    raise ValueError(f"func {name} not found")


def must_loop(arr: t.Iterable, func: t.Callable, custom_error):
  """raise custom error if not looped"""
  is_loop, set_loop = looped()

  for element in arr:
    func(element)
    set_loop()

  if not is_loop():
    raise custom_error("must loop")


def looped():
  loop = False

  def is_loop():
    return loop

  def set_loop():
    nonlocal loop
    loop = True

  return is_loop, set_loop


Result = t.TypeVar("Result")


def try_again(
  total: int,
  retry_f: t.Callable[[], Result],
  is_validate_func: t.Callable[[t.Any], bool],
  fail_message: str,
) -> Result:
  i = 0
  res = retry_f()
  while not is_validate_func(res):
    res = retry_f()
    i += 1
    if i == total:
      raise Exception(f"Cannot find valid result, {fail_message}")

  return res



from PIL import Image
import streamlit as st
import pandas as pd
import pathlib
import numpy as np
from meterviewer.datasets import imgv
from matplotlib import pyplot as plt


def cmd():
  st.Page("compare page", title="Tool to compare images")
  root_path = pathlib.Path("/home/xiuhao/work/Dataset/HRCdata")
  dataset_1, dataset_2 = "Test", "Train"
  num = 1
  x1_name, x2_name = "x_all.npy", "x_train.npy"
  x1_path = root_path / dataset_1 / x1_name
  x2_path = root_path / dataset_2 / x2_name
  x1 = np.load(x1_path)
  x2: np.ndarray = np.load(x2_path)

  def save_img(path, arr):
    im = Image.fromarray(arr)
    im.save(path)

  im = x1[num]
  save_img("/tmp/im1.png", im)
  print("example saved.")

  inx = imgv.find_images(x2)
  if inx == -1:
    print("No same image found.")
  else:
    print(inx)
    print("Same image found.")
    save_img("/tmp/im2.png", x2[inx])


def main():
  st.title("Tool to compare images.")
  root_path = pathlib.Path("/home/xiuhao/work/Dataset/HRCdata")
  root_path = st.text_input("Enter the root path", value=str(root_path))
  root_path = pathlib.Path(root_path)

  dataset_1 = st.text_input("Enter the dataset 1 name", value="Test")
  dataset_2 = st.text_input("Enter the dataset 2 name", value="Train")

  num = st.text_input("Enter the number of data")
  x1_name = st.text_input("Enter the value of x1_name", value="x_all.npy")
  # y_name = st.text_input("Enter the value of y_name", value="y_test.npy")
  x2_name = st.text_input("Enter the value of x2_name", value="x_train.npy")

  try:
    data_load_state = st.text("reading data...")
    num = int(num)
    data_load_state.text("Done!")
  except ValueError:
    data_load_state.text("Please enter a number")
    return

  # folder_path = root_path / path
  x1_path = root_path / dataset_1 / x1_name
  x2_path = root_path / dataset_2 / x2_name

  assert x1_path.exists()
  assert x2_path.exists()

  x1 = np.load(x1_path)
  x2: np.ndarray = np.load(x2_path)

  st.text(f"x1_shape: {x1.shape}, x2_shape: {x2.shape}")

  st.text("Selected image: ")
  st.image(x1[num], caption=f"Meterdata {num}")

  # raise Exception("Not implemented yet")

  im = x1[num]
  if st.button("Search"):
    state = st.text("searching")
    inx = imgv.find_images(im, x2.tolist())
    if inx == -1:
      st.text("No same image found.")
    else:
      st.text("Same image found.")
      st.image(x2[inx], caption=f"Meterdata {inx}")
    state.text("finished")


if __name__ == "__main__":
  # cmd()
  main()



"""view the dataset"""

import pathlib
import random
import typing as t

import cv2

from meterviewer.generator import jsondb
from meterviewer.img import draw
from meterviewer.types import Rect


# 从数据集中随机获取一张图
def get_random_images(
  dataset: str,
  digit_num: int,
  stage: str,
  get_base_dir=jsondb.get_base_dir,
):
  images = jsondb.get_images_with_full_path(
    get_base_dir(),
    dataset,
    digit_num,
    is_test=stage == "test",
  )
  return images


# 生成样例数据图片列表的函数
def gen_plt_images(datasets: jsondb.DatasetList, digit_num: int, stage: str):
  # 输入 get_random_image 的函数
  def fn(get_random_image=get_random_image):
    plt_images = []
    for dataset in datasets.dataset_list:
      images = get_random_images(dataset, digit_num, stage)
      if len(images) == 0:
        print(f"empty dataset: {dataset}")
        continue

      item = {
        "dataset": (dataset, len(images)),
        "image": get_random_image(images),
      }
      plt_images.append(item)
    return plt_images

  return fn


# 根据数据集随机返回一张图片
def get_random_image_by_dataset(dataset: str, digit_num: int, stage: str):
  images = get_random_images(dataset, digit_num, stage)
  if len(images) == 0:
    raise ValueError(f"empty dataset: {dataset}")

  def fn(image: pathlib.Path):
    area = jsondb.read_area_pos(image)
    area = Rect.model_validate(area)
    return area

  return get_random_image(images, read_image_area=fn)


# 随机抽取一张图片，画框
def get_random_image(
  images: list[str], read_image_area: t.Callable[[pathlib.Path], Rect]
):
  """
  draw random image from image list
  images: image list, full path
  """
  image = pathlib.Path(random.choice(images))
  area = read_image_area(image)
  im = cv2.imread(image)
  drawed_im = draw.draw_rectangle(im, area)
  print(area)
  return drawed_im


class DatasetView(object):
  """DatasetView"""

  def get_random_image_by_dataset(self, dataset: str, digit_num: int, stage: str):
    images = get_random_images(dataset, digit_num, stage, self.get_base_dir)
    self.check_images(images, dataset)
    return get_random_image(images, self.read_image_area)

  def check_images(self, images: list[str], dataset: str):
    if len(images) == 0:
      raise ValueError(f"empty dataset: {dataset}")

  def get_base_dir(self):
    return jsondb.get_base_dir()

  def read_image_area(self, image: pathlib.Path):
    area = jsondb.read_area_pos(image)
    area = Rect.model_validate(area)
    return area



from __future__ import annotations

import pathlib
import typing as t

from meterviewer import files
from meterviewer import types as T
from meterviewer.datasets import dataset


def view_merge_np(
  current_dataset: str | pathlib.Path,
  view_dataset: t.Callable[[int, T.ImgList], None] = dataset.view_dataset,
  get_x_y: T.NameFunc = lambda: (T.x_name, T.y_name),
):
  """查看已处理的数据集

  Args:
      current_dataset: 数据集路径
      view_dataset: 用于查看数据集的回调函数
      get_x_y: 获取X和Y文件名的函数
  """
  pp = pathlib.Path(current_dataset)

  x_name, _ = get_x_y()
  view_func = dataset.view_dataset_on_disk(x_name)

  view_func(
    prefix_name=pp,
    view_dataset=view_dataset,
    load_from_disk=files.load_from_disk,
  )


def read_details(current_dataset: str) -> t.Optional[t.Dict]:
  """读取数据集的详细信息

  Args:
      current_dataset: 数据集路径

  Returns:
      包含数据集详细信息的字典，如果文件不存在则返回None
  """
  return files.read_toml(pathlib.Path(current_dataset) / "details.gen.toml")


def get_x_y_name() -> t.Tuple[str, str]:
  """获取默认的X和Y文件名

  Returns:
      包含X和Y文件名的元组
  """
  return T.x_name, T.y_name


def write_details(
  current_dataset: str | pathlib.Path,
  get_xy_name: T.NameFunc = get_x_y_name,
):
  """写入数据集的详细信息

  Args:
      current_dataset: 数据集路径
      get_xy_name: 获取X和Y文件名的函数
  """
  pp = pathlib.Path(current_dataset)

  def write_to_file(details, overwrite=True):
    p = pp / "details.gen.toml"
    if not overwrite and p.exists():
      print("Failed to write, file exists")
      return
    return files.write_toml(p, details)

  x_name, y_name = get_xy_name()

  dataset.show_details(
    get_x_train=lambda: files.load_from_disk(pp / x_name),
    get_y_train=lambda: files.load_from_disk(pp / y_name),
    get_details=lambda x, y: dataset.get_details(pp, x, y),
    write_to_file=write_to_file,
  )


class NPView(object):
  def view(self, current_dataset: str | pathlib.Path, default_way=1):
    if default_way == 1:
      view_func = self.view_dataset
    else:
      view_func = self.view_dataset_in_rows

    return view_merge_np(
      current_dataset=current_dataset,
      view_dataset=view_func,
      get_x_y=self.get_x_y_name,
    )

  def view_dataset(self, num: int, imglist: T.ImgList):
    return dataset.view_dataset(num, imglist)

  def view_dataset_in_rows(self, num: int, imglist: T.ImgList):
    return dataset.view_dataset_in_rows(num, imglist)

  def get_x_y_name(self):
    return T.x_name, T.y_name



import os
import pathlib

from meterviewer.generator.jsondb import get_base_dir, get_random_data


def test_get_random_data(set_config):
  data = get_random_data(is_relative_path=False)
  assert pathlib.Path(data).exists()

  data_path = get_random_data(is_relative_path=True)
  assert os.path.exists(get_base_dir() / data_path)



"""create jsondb for meter-viewer"""

import glob
import json
import os
import pathlib
import random
import typing as t
from functools import lru_cache

import toml
from pydantic import BaseModel

from meterviewer.datasets.read.detection import read_area_pos

from .base import Generator
from .schema import Item, MeterDB


# cache config function (only read disk once), returns get_random_dataset and load_conf
def load_config(config_path: pathlib.Path) -> t.Callable[[], dict]:
  data: t.Optional[dict] = None

  def load_conf() -> dict:
    nonlocal data
    if data is None:
      with open(config_path, "r") as f:
        data = toml.load(f)
    assert data is not None, (config_path, data)
    return data

  return load_conf


get_local_config = None


# 随机选择一个数据集
def get_random_dataset(is_train: bool = True) -> str:
  dataset_list = get_dataset(is_train)
  return random.choice(dataset_list)


class DatasetList(BaseModel):
  digit_num: int
  dataset_list: list[str]


def _get_dataset(config, digit_number: int, is_train: bool = True) -> DatasetList:
  if is_train:
    key = "train_dataset"
  else:
    key = "test_dataset"

  if digit_number not in [5, 6]:
    raise ValueError(f"digit_number must be 5 or 6, but got {digit_number}")
  return DatasetList(
    digit_num=digit_number, dataset_list=config["base"][f"{digit_number}_digit"][key]
  )


# 获取数据集列表
def get_dataset(digit_number: int, is_train: bool = True) -> DatasetList:
  config = get_local_config()
  return _get_dataset(config, digit_number, is_train)


def _get_base_dir(config):
  return config["base"]["root_path"]


@lru_cache
def get_base_dir() -> str:
  """获取数据集的 base_dir"""
  config = get_local_config()
  return _get_base_dir(config)


@lru_cache
def get_mid_path(digit_num: int, is_test: bool) -> str:
  """获取数据集的 mid_path"""
  if digit_num not in [5, 6]:
    raise ValueError(f"digit_num must be 5 or 6, but got {digit_num}")

  if digit_num == 5:
    if is_test:
      return "lens_5/XL/DATA"
    else:
      return "lens_5/CS/Data_cs"
  if digit_num == 6:
    if is_test:
      return "lens_6/CS/all_CS"
    else:
      return "lens_6/XL/XL"
  raise ValueError(f"digit_num must be 5 or 6, but got {digit_num}")


def get_random_data(
  is_test: bool = False,
  is_relative_path: bool = True,
) -> pathlib.Path:
  """随机获取一个数据集下的图片"""
  dataset = get_random_dataset(is_train=not is_test)
  base_dir = get_base_dir()
  mid_path = get_mid_path(is_test=is_test)

  data_path = glob.glob(str(pathlib.Path(base_dir) / mid_path / dataset / "*.jpg"))
  random_path = random.choice(data_path)
  if is_relative_path:
    return pathlib.Path(random_path).relative_to(base_dir)
  else:
    return pathlib.Path(random_path)


# Not a pure function.
def set_local_config(infile: pathlib.Path):
  """设置本地配置, 只针对直接调用的函数有效"""
  global get_local_config
  get_local_config = load_config(config_path=infile)


def get_images_with_full_path(
  base_dir: str,
  dataset_name: str,
  digit_number: int,
  is_test: bool,
) -> list[str]:
  full_path = str(
    pathlib.Path(base_dir).expanduser()
    / get_mid_path(digit_number, is_test)
    / dataset_name
    / "*.jpg"
  )
  print("full_path: ", full_path)
  data_path = glob.glob(full_path)
  return data_path


def get_dataset_dir(
  digit_number: int,
  is_test: bool,
) -> pathlib.Path:
  return pathlib.Path(get_base_dir()) / get_mid_path(digit_number, is_test)


def gen_db(
  infile: pathlib.Path,
  output: pathlib.Path,
  stage: t.Literal["train", "test"],
):
  """
  读取数据集下所有的图片, 以及点的位置, 生成一个json文件
  """

  # set the get_local_config
  set_local_config(infile)

  data = []
  is_train = stage == "train"
  is_test = stage == "test"

  base_dir = get_base_dir()

  # 获取 5-6 位的大型数据集
  for digit in [6, 5]:
    dataset = get_dataset(digit, is_train)
    # 遍历小数据集，根据不同的仪表型号区分的

    for dataset_name in dataset.dataset_list:
      # 获取所有的图像路径
      img_paths = get_images_with_full_path(
        base_dir,
        dataset_name,
        dataset.digit_num,
        is_test,
      )

      if not img_paths:
        print(f"no image found for dataset: {dataset_name}, {img_paths}")
        continue

      for img_path in img_paths:
        assert os.path.exists(img_path), f"image not found: {img_path}"
        assert os.path.isfile(img_path), f"image is not a file: {img_path}"

        # 读取图像中的点
        rect = read_area_pos(pathlib.Path(img_path))

        # 使用相对路径可以避免生成的 db 无法在其他机器上使用
        relative_path = pathlib.Path(img_path).relative_to(base_dir)

        item = Item(
          filepath=str(relative_path),
          dataset=dataset_name,
          xmin=rect["xmin"],
          xmax=rect["xmax"],
          ymin=rect["ymin"],
          ymax=rect["ymax"],
        )
        data.append(item)

  meter_db = MeterDB(data=data)

  with open(output, "w") as f:
    json.dump(meter_db.model_dump(), f)

  return output


class JSONDB(Generator):
  """generate jsondb, inherit get_local_config method to get local config"""

  def get_infile(self):
    raise NotImplementedError("get_infile")

  def get_local_config(self) -> dict:
    fn = load_config(config_path=self.get_infile())
    return fn()

  def get_dataset(self, digit_number: int, is_train: bool = True) -> DatasetList:
    return _get_dataset(self.get_local_config(), digit_number, is_train)

  def get_base_dir(self) -> str:
    """获取数据集的 base_dir"""
    config = self.get_local_config()
    return config["base"]["root_path"]



"""view image and value, process md5 value."""

import pathlib
import typing as t
from pathlib import Path as P

import numpy as np
from loguru import logger
from matplotlib import pyplot as plt
from PIL import ImageFile
from tqdm import tqdm

from meterviewer import T
from meterviewer.datasets.read import config
from meterviewer.img import cmp

logger.add("./logs/meterviewer-proc.log")

# 修复 PIL 无法读取截断图片的问题; matplotlib 底层居然是 PIL？是我搞错了吗？
ImageFile.LOAD_TRUNCATED_IMAGES = True


def view_one_img_v(filepath: P) -> t.Tuple[T.NpImage, str, T.Rect]:
  """同时浏览图片和值"""
  im = plt.imread(str(filepath))
  v, rect = config.get_xml_config(filepath)
  return (im, v, rect)


hash_store = {}


def save_hash(im: T.NpImage, index: int):
  """save image file hash value."""
  hash = cmp.get_hash(im)
  res = hash_store.get(hash, None)
  if res is None:
    hash_store[hash] = index
  else:
    logger.error(f"Hash: {hash} is same with {res} and {index}")
    raise Exception("Img should not be same.")


def find_images(img_list: T.ImgList) -> int:
  logger.debug("start finding...")
  for i, img in tqdm(enumerate(img_list)):
    try:
      save_hash(img, i)
    except Exception as e:
      logger.warning(f"warning: {e} in {i}")
      return i
  # index in the np_path
  return -1


def find_hash_in_numpy(im: T.NpImage, np_path: pathlib.Path) -> int:
  save_hash(im, -1)
  x = np.load(np_path)
  res = find_images(list(x))
  return res



import pytest
from meterviewer import func as F


def test_try_again():
  assert F.try_again(5, lambda: 1, lambda x: x == 1, fail_message="cannot get 1") == 1

  with pytest.raises(Exception):
    assert F.try_again(5, lambda: 1, lambda x: x == 2, fail_message="cannot get 1") == 1


def test_must_loop():
  def func(x):
    assert x == 1

  class LoopError(Exception):
    pass

  F.must_loop([1, 1, 1], func, LoopError)

  assert pytest.raises(LoopError, F.must_loop, [], func, LoopError)



from string import Template


def dataset():
  data = {}
  str_template = Template("$name_$peroid.npy")

  names = ["x", "y"]
  peroids = ["train", "test"]

  s = str_template.safe_substitute(name="train", peroid="2021")

  def get_root_path():
    pass



from meterviewer import F
from meterviewer.datasets import dataset
from meterviewer.datasets.read import config


def test_read_single_digit(root_path):
  p = config.read_single_digit_rect(
    root_path / r"lens_6/XL/XL/M1L1XL" / "config" / "res.xml"
  )

  assert len(p) == 6
  p: list[config.RectO]
  assert p[0].xmax == "105", p[0]


def test_exist_dataset(root_path):
  assert dataset.get_dataset_path(root_path, r"M1L3XL").exists()


def test_read_xml(root_path):
  v, r = config.read_xml_to_get(
    root_path / r"lens_6/XL/XL/M1L3XL" / "baocun" / "2018-11-23-12-16-01.xml",
    config.read_rect_from_node,
  )
  assert v == "000994"
  assert r.get("xmin") == "79"


def test_img_to_xml(root_path):
  path_list = [
    config.get_xml_config_path(
      root_path / r"lens_6/XL/XL/M1L3XL" / "2018-11-23-12-16-01.jpg"
    ),
    config.get_xml_config_path(
      root_path / r"lens_6/XL/XL/M1L3XL" / "2018-11-23-12-16-01.jpg", "single"
    ),
  ]

  def path_exist(p):
    assert p.exists(), p

  F.must_loop(path_list, path_exist, custom_error=Exception)



"""generate sqlite3 database"""

import typing as t
from pathlib import Path as P

from meterviewer import T, files
from meterviewer.datasets import dataset, imgv
from meterviewer.datasets.read import single
from meterviewer.models import littledb
from meterviewer.values import is_carry

from .base import Generator

# Type hint for database insertion function
dbInsertFunc = t.Callable[[str, int, bool], None]


def generate_for_one_dataset(dataset: P, insert: dbInsertFunc):
  """
  Process a single dataset and insert its image data into the database.

  Args:
      dataset (Path): Absolute path to the dataset directory
      insert (dbInsertFunc): Callback function to insert data into the database
          Expected signature: (image_path: str, value: int, is_carry: bool) -> None

  Raises:
      AssertionError: If dataset path is not absolute
  """
  assert dataset.is_absolute()
  for img_file in files.scan_pics(dataset):
    _, v, _ = imgv.view_one_img_v(img_file)
    # transform str to int
    val = int(v)

    insert(str(img_file), val, is_carry(v))


def generate_db_for_all(root: P, db_path: P):
  """
  Generate a single database containing data from all datasets.

  Args:
      root (Path): Root directory containing all datasets
      db_path (Path): Relative path where the database should be created

  Raises:
      AssertionError: If db_path is absolute
  """
  # insert to one database.
  assert not db_path.is_absolute()

  insert, _ = littledb.create_db(str(db_path))

  def generate_one(dataset: P):
    return generate_for_one_dataset(single.get_dataset_path(root, str(dataset)), insert)

  dataset.handle_datasets(
    root,
    generate_one,
  )


def create_db(root_path: P):
  """
  Create separate databases for each dataset in the root directory.
  Each database will be named 'items.db' and placed in its respective dataset directory.

  Args:
      root_path (Path): Root directory containing all datasets
  """
  db_name = "items.db"

  def handle_dataset(dataset_name: P):
    dataset_path = dataset.get_dataset_path(root_path, str(dataset_name))
    littledb.create_db(str(dataset_path / db_name))

  dataset.handle_datasets(
    root_path,
    handle_func=handle_dataset,
  )


class DB(Generator):
  """generate sqlite3 database"""

  def load_root_path(self):
    raise NotImplementedError("load_root_path")

  def create_db(self):
    return create_db(self.load_root_path())

  def generate_db_for_all(self, db_path: P):
    return generate_db_for_all(self.load_root_path(), db_path)

  def generate_for_one_dataset(self, dataset: P, insert: dbInsertFunc):
    return generate_for_one_dataset(dataset, insert)



# rotate image.
import os
import pathlib
import typing as t

import numpy as np
import streamlit as st

from meterviewer.config import get_root_path
from tmd.dataset import NumpyDataset

try:
  import torch
  from torchvision.transforms import v2
except ImportError:
  st.text("Please install torch and torchvision.")

torch.manual_seed(17)


def trans_function():
  trans = torch.nn.Sequential(
    v2.Pad(10, fill=0, padding_mode="constant"),
    v2.RandomRotation(30),
    # v2.RandomResizedCrop(size=(224, 224), antialias=True),
    v2.RandomHorizontalFlip(p=0.5),
    # transforms.CenterCrop(10),
    v2.ColorJitter(),
  )
  return trans


def rotate():
  st.title("rotate the image.")
  root_path = get_root_path()
  root_path = st.text_input("Enter the root path", value=str(root_path))
  root_path = pathlib.Path(root_path)

  path = st.text_input("Enter the path of the data", value="generated_merged")
  st.text(os.listdir(root_path / path))

  num = st.text_input("Enter the number of data")
  x_name = st.text_input("Enter the value of x_name", value="x_test.npy")
  y_name = st.text_input("Enter the value of y_name", value="y_test.npy")

  try:
    num_v: int = int(num)
  except ValueError:
    st.text("Please enter a number")

  folder_path = root_path / path
  x_path = folder_path / x_name
  y_path = folder_path / y_name

  status = st.text("loading dataset...")
  ds = NumpyDataset(
    folder_path,
    x_path,
    y_path,
    transform=trans_function(),
  )
  status.text("Done.")

  if st.button("Transform"):
    x, y = ds[num_v]
    st.text(f"x_shape: {x.size}, y_shape: {y.shape}")

    st.image(x, caption=f"Meterdata {num}")
    st.text(f"Meterdata {num} is {y}")


if __name__ == "__main__":
  rotate()



from meterviewer.models import littledb
from meterviewer.datasets import dataset
from pathlib import Path as P


def test_insert_one(root_path, remove_files):
  dataset_name = "M1L1XL"
  p = dataset.get_dataset_path(root_path, dataset_name)
  db = p / "items-temp.db"
  insert_one, _ = littledb.create_db(str(db))
  assert db.exists()

  insert_one("test", 0, False)

  # seems not work under windows
  remove_files(db)
  # db.unlink()


def test_create_db(root_path, remove_files):
  p = P("test.db")
  littledb.create_db(str(p))
  remove_files(p)

  p = P(root_path / "test.db")
  littledb.create_db(str(p))
  remove_files(p)



# 处理图像以适应训练要求

from __future__ import annotations

import sys
import typing as t

import numpy as np
from matplotlib import pyplot as plt

# from matplotlib import pyplot as plt
from .. import types as T
from .resize import check_img_size, resize_img, resize_imglist, size_check  # noqa


def np_to_img(data: np.ndarray) -> T.ImgList:
  """将numpy数组转换为图像列表"""
  return list(data)


def join_img(
  imglist: T.ImgList,
  check_func: t.Callable[[t.Any], t.Any],
) -> T.NpImage:
  """水平合并图像
  Args:
      imglist: 图像列表
      check_func: 检查函数
  Returns:
      合并后的numpy图像数组
  """
  check_func(imglist)
  return np.hstack(imglist)


def get_random_img(num: int, img_from: t.Callable) -> T.NpImage:
  """获取随机图像
  Args:
      num: 图像的数字位数
      img_from: 图像来源函数
  Returns:
      随机生成的numpy图像数组
  """
  get_img = img_from()
  return get_img(num)


def img_from(folder: str = ""):
  """从文件夹获取所有图像
  Args:
      folder: 图像文件夹路径
  Returns:
      获取图像的函数
  """

  def get_img(num):
    # 生成随机图像数组(10x20)，像素值范围1-255
    return np.random.randint(1, 255, size=(10, 20))

  return get_img


def get_img_list(nums: t.List[int]) -> t.List[T.NpImage]:
  """根据数字列表获取对应的图像列表
  Args:
      nums: 数字列表
  Returns:
      图像列表
  """
  imgs = []
  for i in nums:
    imgs.append(get_random_img(int(i), lambda: None))
  return imgs


def number_to_string(number: int, length: int) -> t.List[str]:
  """将数字转换为固定长度的字符串列表
  Args:
      number: 要转换的数字
      length: 期望的字符串长度
  Returns:
      字符串列表，不足位数用0填充
  """
  return list(str(number).zfill(length))


def empty_check(*args, **kwargs):
  """空检查函数，不执行任何操作"""
  pass


def gen_block_img(number: int, length: int):
  """在内存中生成图像块，用作示例函数
  Args:
      number: 要生成图像的数字
      length: 期望的长度
  Returns:
      合并后的图像块
  """
  num_l = [int(i) for i in number_to_string(number, length)]
  return join_img(get_img_list(num_l), empty_check)


def show_img(img, is_stop):
  """显示图像
  Args:
      img: 要显示的图像
      is_stop: 是否在显示后退出程序
  """
  plt.imshow(img)
  plt.show()
  if is_stop:
    sys.exit(-1)


def gen_empty_im(size: t.Tuple[int, int, int]):
  """生成指定大小的空白图像
  Args:
      size: 图像大小元组 (高度, 宽度, 通道数)
  Returns:
      空白图像数组
  """
  return np.zeros(size, dtype=np.uint8)



# import cv2
from meterviewer import types as T

PartImg = T.NpImage  # reading area of meter image


def cut_img(img: T.NpImage, rect: T.Rect) -> PartImg:
  """根据 rect 裁剪 img"""
  cropped_image = img[rect.ymin : rect.ymax, rect.xmin : rect.xmax]
  return cropped_image



import pathlib
import platform
import typing as t


def get_root_path():
  def get_system_type() -> t.Literal["macOS", "Windows", "Other", "Linux"]:
    """
    判断当前系统类型是 macOS 还是 Windows。

    Returns:
        str: 'macOS', 'Windows' 或 'Other'，分别表示 macOS 系统、Windows 系统或其他系统。
    """
    system = platform.system()

    if system == "Darwin":
      return "macOS"
    elif system == "Windows":
      return "Windows"
    elif system == "Linux":
      return "Linux"
    else:
      return "Other"

  if get_system_type() == "macOS":
    # expand userhome
    return pathlib.Path("/Volumes/Fast SSD/Research/Dataset/MeterData")
  elif get_system_type() == "Windows":
    return pathlib.Path(r"D:\Store\MeterData")
  elif get_system_type() == "Linux":
    return pathlib.Path("~/Work/Dataset/MeterData").expanduser()
  else:
    raise Exception("Not work under Linux now. 2024-05-21.")



"""
test dataset functions.
"""

from __future__ import annotations

import functools
import pathlib

import pytest

from meterviewer import files
from meterviewer import types as T
from meterviewer.datasets import dataset
from meterviewer.datasets.read import single
from meterviewer.img import process
from tests.utils import show_img


def test_create_label():
  assert dataset.create_str_with_fill(1234, 5, 8) == [
    "0",
    "1",
    "2",
    "3",
    "4",
    "x",
    "x",
    "x",
  ]
  assert dataset.create_str_with_fill(12340, 5, 8) == [
    "1",
    "2",
    "3",
    "4",
    "0",
    "x",
    "x",
    "x",
  ]


def test_python_syntax():
  arr = ["1", "2", "3", "4"]
  arr.extend(["x"] * 3)
  assert arr == list("1234xxx")
  assert arr == ["1", "2", "3", "4", "x", "x", "x"]

  arr = list("1234")
  arr.extend(["x"] * 3)
  assert arr == list("1234xxx")


def test_fill_digit():
  assert dataset.fill_digit(list("12312"), 7) == list("12312xx")
  assert dataset.fill_digit(list("01234"), 8) == list("01234xxx")
  assert dataset.fill_digit(["0", "1", "2", "3", "4"], 8) == [
    "0",
    "1",
    "2",
    "3",
    "4",
    "x",
    "x",
    "x",
  ]
  assert list("12312xx") == ["1", "2", "3", "1", "2", "x", "x"]
  assert list("1212xx") == ["1", "2", "1", "2", "x", "x"]
  assert list("0123") == ["0", "1", "2", "3"]

  with pytest.raises(ValueError):
    dataset.fill_digit(list("12312"), 4)
    dataset.fill_digit(list("12"), 4)


def test_view_on_disk(root_path):
  f = dataset.view_dataset_on_disk(T.x_name)
  f(
    prefix_name=root_path / r"generated",
    load_from_disk=files.load_from_disk,
    view_dataset=dataset.view_dataset,
    show=True,
  )


def test_create_dataset(root_path):
  P = functools.partial

  def read_rand_img(digit: int | str):
    return single.read_rand_img(
      digit=digit,
      root=root_path,
      get_dataset=lambda: "M1L1XL",
    )

  gen_block = P(
    dataset.generate_block_img,
    join_func=dataset.join_with_resize,
    read_rand_img=read_rand_img,
  )

  path = root_path / pathlib.Path(r"generated")
  path.mkdir(exist_ok=True)

  filesave = P(
    files.save_img_labels_with_default,
    prefix_name=path,
    save_to_disk=files.save_to_disk,
  )

  def check_imgs(imglist):
    size = imglist[0].shape
    show_img(imglist[0])
    imgs = process.resize_imglist(imglist)
    for im in imgs:
      show_img(im)
      assert size == im.shape

  fn = dataset.create_dataset_func(check_imgs=lambda x: None, total=9)
  imgs, labels = fn(length=5, nums=10, gen_block_img=gen_block)
  filesave(imgs=imgs, labels=labels)


def test_generate_block_img(root_path):
  def read_rand_img(digit: int | str):
    return single.read_rand_img(
      digit=digit,
      root=root_path,
      get_dataset=lambda: "M1L1XL",
    )

  im = dataset.generate_block_img(
    ["1", "2", "3", "4"],
    process.join_img,
    read_rand_img,
  )
  show_img(im)

  im = dataset.generate_block_img(
    ["1", "2", "3", "5", "6"],
    dataset.join_with_resize,
    read_rand_img,
  )
  show_img(im)

  im = dataset.generate_block_img(
    ["1", "2", "3", "5", "6", "x", "x"],
    dataset.join_with_resize,
    read_rand_img,
  )
  show_img(im)


def test_get_random_dataset(root_path):
  _, index = dataset.get_random_dataset(root_path, dataset.get_dataset_list)
  assert index in range(0, 74)


def test_read_random_img(root_path):
  im = single.read_rand_img(root_path, lambda: "M1L1XL", 5)
  show_img(im)

  im = single.read_rand_img(root_path, lambda: "M1L1XL", "x")
  show_img(im)


def test_read_random_digit(root_path):
  path_gen = single.read_single_digit(root_path, lambda: "M1L1XL", 0, promise=False)()
  p = next(path_gen)
  assert pathlib.Path(p).exists()


def test_dataset_list(root_path):
  # path = root_path / pathlib.Path("lens_6/XL/XL")
  count = 0
  pics = []

  def update_count(name):
    nonlocal count, pics
    pics.append(name)
    count += 1

  dataset.handle_datasets(root_path, update_count)
  assert count == 74
  # assert False, pics



from pathlib import Path as _p

import pytest

from meterviewer.config import get_root_path
from meterviewer.generator import jsondb


def test_path():
  basic_path = get_root_path()
  assert basic_path.exists()

  def full_path(digit_num: int, is_test: bool):
    return basic_path / jsondb.get_mid_path(digit_num, is_test)

  for num, is_test in [(6, True), (6, False), (5, True), (5, False)]:
    f = full_path(num, is_test)
    assert full_path(num, True) != full_path(num, False), "test and train should be different"
    assert f.exists(), f"{f} does not exist."

  with pytest.raises(ValueError):
    full_path(7, True)
